from .AbstractSamplingPerturbationMechanism import AbstractSamplingPerturbationMechanism
import numpy as np
import numpy.random as npr
from scipy.linalg import sqrtm
import multiprocessing as mp
from typing import List
import time

'''
# SanText mechanism
'''

class SanText(AbstractSamplingPerturbationMechanism):
    '''
    # SanText Mechansim
    
    BibTeX of SanText Mechanism, extends CMP mechanism class of the pypanter package:

    @inproceedings{ACL21/YueDu21,
    author    = {Xiang Yue and Minxin Du and Tianhao Wang and Yaliang Li and Huan Sun and Sherman S. M. Chow},
    title     = {Differential Privacy for Text Analytics via Natural Text Sanitization},
    booktitle = {Findings, {ACL-IJCNLP} 2021},
    year      = {2021},
    }
    '''
    def __init__(self, kwargs: dict[str:object]) -> None:
        '''
        Initialization of the SanText Object

        : param kwargs: dict the dictionary containing the parameters of the Mechanism Object 
                        + the specific parameters of the SanText Mechanism ()

        Once the SanText Object is created, the user can use the obfuscateText method to obfuscate 
        the text of the provided text.

        The attributes of the Mechanism Object are:
        - vocab: the vocabulary object containing the embeddings
        - embMatrix: the matrix containing the embeddings
        - index2word: the dictionary containing the index to word mapping
        - word2index: the dictionary containing the word to index mapping
        - epsilon: the epsilon parameter of the mechanism
        - 

        Usage example:
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1 #anyvalue of epsilon must be greater than 0
        >>> lam: float = 0.1 #anyvalue of lambda must be between 0 and 1
        >>> mech1 = SanText({'embPath': embPath, 'epsilon': eps, })
        '''
        super().__init__(kwargs)
        self.name:str = 'SanText'
        self.prob_matrix = {}

    def processText(self, text:List[str])->str:
        '''
        method selfProcessQueryText: this method is used to process the Text and return the obfuscated Text

        : param embs: np.array the embeddings of the words
        : return: str the obfuscated Text
        '''
        length:int = len(text) #number of words in the text
        text_tuple:tuple = tuple(text)
        #get embeddings
        if (text_tuple, self.epsilon) not in self.prob_matrix.keys():
            embs = []
            for word in text:
                if word not in self.vocab.embeddings.keys():
                    embs.append(np.zeros(self.embMatrix.shape[1]) + npr.normal(0, 1, self.embMatrix.shape[1]))
                else:
                    embs.append(self.vocab.embeddings[word])
            embs:np.array = np.array(embs)
        # Compute distance matrix between word and all words in vocabulary
            distance:np.array = self.euclideanDistance(embs, self.embMatrix)
            # Compute probability
            exp_neg_distance:np.array = np.exp(-0.5 * self.epsilon * distance)
            total_sum:np.array = np.reciprocal(np.sum(exp_neg_distance, axis=1))
            probabilities:np.array = exp_neg_distance * total_sum[:, np.newaxis]
            
            # Normalize
            probabilities /= np.sum(probabilities, axis=1, keepdims=True)
            
            # Add the probabilities to the dictionary
            self.prob_matrix[(text_tuple, self.epsilon)] = probabilities
        else:
            probabilities = self.prob_matrix[(text_tuple, self.epsilon)]
        
        # Sampled obfuscated word should be a list of length len(text)
        sampled_indices:List[np.array] = [np.random.choice(self.embMatrix.shape[0], p=prob) for prob in probabilities]
        sampled_obfuscated_text:List[str] = [list(self.vocab.embeddings.keys())[index] for index in sampled_indices]
        
        return " ".join(sampled_obfuscated_text)
        
        


from .AbstractEmbeddingPerturbationMechanism import AbstractEmbeddingPerturbationMechanism
import numpy as np
import numpy.random as npr
from scipy.linalg import sqrtm
from typing import List

'''
# Mahalanobis Mechanism
'''

class Mahalanobis(AbstractEmbeddingPerturbationMechanism):
    '''
    # Mahalanobis
    
    BibTeX of Mahalanobis Mechanism, extends Mechanism mechanism class of the pypantera package:

    @inproceedings{xu-etal-2020-differentially,
    title = "A Differentially Private Text Perturbation Method Using Regularized Mahalanobis Metric",
    author = "Xu, Zekun and Aggarwal, Abhinav and Feyisetan, Oluwaseyi and Teissier, Nathanael",
    editor = "Feyisetan, Oluwaseyi and Ghanavati, Sepideh  and Malmasi, Shervin and Thaine, Patricia",
    booktitle = "Proceedings of the Second Workshop on Privacy in NLP",
    month = nov,
    year = "2020",
    address = "Online",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/2020.privatenlp-1.2.pdf",
    doi = "10.18653/v1/2020.privatenlp-1.2",
    pages = "7--17",
    abstract = "Balancing the privacy-utility tradeoff is a crucial requirement of many practical machine learning systems that deal with sensitive customer data. A popular approach for privacy- preserving text analysis is noise injection, in which text data is first mapped into a continuous embedding space, perturbed by sampling a spherical noise from an appropriate distribution, and then projected back to the discrete vocabulary space. While this allows the perturbation to admit the required metric differential privacy, often the utility of downstream tasks modeled on this perturbed data is low because the spherical noise does not account for the variability in the density around different words in the embedding space. In particular, words in a sparse region are likely unchanged even when the noise scale is large. In this paper, we propose a text perturbation mechanism based on a carefully designed regularized variant of the Mahalanobis metric to overcome this problem. For any given noise scale, this metric adds an elliptical noise to account for the covariance structure in the embedding space. This heterogeneity in the noise scale along different directions helps ensure that the words in the sparse region have sufficient likelihood of replacement without sacrificing the overall utility. We provide a text-perturbation algorithm based on this metric and formally prove its privacy guarantees. Additionally, we empirically show that our mechanism improves the privacy statistics to achieve the same level of utility as compared to the state-of-the-art Laplace mechanism.",
    }
    '''
    def __init__(self, kwargs: dict[str:object]) -> None:

        '''
        Initialization of the Mahalanobis Object

        : param kwargs: dict the dictionary containing the parameters of the Mechanism Object 
                        + the specific parameters of the Mahalanobis Mechanism (lambda)
        
        Once the Mahalanobis Object is created, the user can use the obfuscateText method to obfuscate 
        the text of the provided text.

        The attributes of the Mechanism Object are:
        - vocab: the vocabulary object containing the embeddings
        - embMatrix: the matrix containing the embeddings
        - index2word: the dictionary containing the index to word mapping
        - word2index: the dictionary containing the word to index mapping
        - epsilon: the epsilon parameter of the mechanism
        - lam: the lambda parameter of the Mahalanobis mechanism
        - sigma_loc: parameter used for pulling noise in obfuscation

        Usage example:
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1 #anyvalue of epsilon must be greater than 0
        >>> lam: float = 0.1 #anyvalue of lambda must be between 0 and 1
        >>> mech1 = Mahalanobis({'embPath': embPath, 'epsilon': eps, 'lambda': lam})
        '''

        super().__init__(kwargs)
        assert 'lam' in kwargs, 'The lambda parameter must be provided'
        assert kwargs['lam'] >= 0 and kwargs['lam'] <= 1, 'The lambda parameter must be between 0 and 1'
        self.lam: float = kwargs['lam']
        cov_mat = np.cov(self.embMatrix.T, ddof=0)
        sigma = cov_mat/ np.mean(np.var(self.embMatrix.T, axis=1))
        self._sigmaLoc = sqrtm(self.lam * sigma + (1 - self.lam) * np.eye(self.embMatrix.shape[1]))   
        self.name:str = 'Mahalanobis'   

    def pullNoise(self) -> np.array:
        '''
        method pullNoise: this method is used to pull noise accordingly 
        to the definition of the Mahalanobis mechanism, see BibTeX ref
        : return: np.array the noise pulled
        Usage example:
        (Considering that the Mechanism Object mech1 has been created 
        as in the example of the __init__ method)
        >>> mech1.pullNoise()
        '''

        N: np.array = npr.multivariate_normal(
            np.zeros(self.embMatrix.shape[1]), 
            np.eye(self.embMatrix.shape[1])
            )
        X: np.array = N / np.sqrt(np.sum(N ** 2))
        X: np.array = np.dot(self._sigmaLoc, X)
        X: np.array = X / np.sqrt(np.sum(X ** 2))
        Y: np.array = npr.gamma(
            self.embMatrix.shape[1], 
            1 / self.epsilon
            )
        Z: np.array = Y * X
        return Z

    def processText(self, 
                     embs: np.array) -> str:
        
        '''
        method processText: this method is used to process the Text and return the obfuscated Text

        : param embs: np.array the embeddings of the words
        : return: str the obfuscated Text

        Usage example:
        (Considering that the Mechanism Object mech1 has been created
        as in the example of the __init__ method)

        # Assuming that the embeddings of the words are known, e.g.: [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        >>> embs: np.array = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        >>> mech1.processText(embs)
        '''
        
        length: int = len(embs)
        distance: np.array = self.euclideanDistance(embs, self.embMatrix)
        closest: np.array = np.argpartition(distance, 1, axis=1)[:, :1]
        finalText: List[str] = []
        for i in range(length):
            finalText.append(list(self.vocab.embeddings.keys())[closest[i][0]])
        return ' '.join(finalText)


from .utils.vocab import Vocab
import numpy as np
import pandas as pd
from typing import List

class AbstractTextObfuscationDPMechanism():
    '''
    Abstract class Mechanism: this class is used to define mechanism
    '''
    def __init__(self, kwargs: dict[str:object]) -> None:
        self.vocab: Vocab = Vocab(kwargs['embPath']) #load the vocabulary
        self.embMatrix: np.array = np.array(
            list(self.vocab.embeddings.values())
            ) #load the embeddings matrix
        self._word2index: dict = {
            word: i 
            for i, word in enumerate(self.vocab.embeddings.keys())
            } 
        self._index2word: dict = {
            i: word for word,i in self._word2index.items() 
        }        
        assert 'epsilon' in kwargs, 'The epsilon parameter must be provided'
        assert kwargs['epsilon'] > 0, 'The epsilon parameter must be greater than 0'
        self.epsilon: float = kwargs['epsilon'] #set the epsilon parameter
    
    def indexes2words(self, indexes:list) -> List[str]:
        return [self._index2word[e] for f in indexes for e in f]
    
    def obfuscateText(self, 
                      data:pd.DataFrame, 
                      numberOfTimes: int) -> List[str]:
        '''
        method obfuscateText: this method is used to obfuscate the text of the provided text 
        it uses the istance specific obfuscationText method
        

        : param data: pd.DataFrame of the texts to obfuscate
        : param numberOfTimes: int the number of times to obfuscate the text
        : return: df the obfuscated texts

        Usage example:
        (Considering that the Mechanism Object mech1 has been created
        as in the example of the __init__ method)

        >>> text: str = 'what is the capitol of france'
        >>> mech1.obfuscateText(text, 1)
        '''
        return self.obfuscateText(data, numberOfTimes)

    @staticmethod
    def euclideanDistance(x: np.array, 
                          y: np.array) -> np.array:
        '''
        method euclideanDistance: this method is used to compute the euclidean distance between two matrices

        Remark: this method is an obtimization of the euclidean distance computation between two matrices

        : param x: np.array the first matrix
        : param y: np.array the second matrix
        : return: np.array the euclidean distance matrix between the two matrices

        Usage example:
        >>> x: np.array = np.array([1, 2, 3])
        >>> y: np.array = np.array([4, 5, 6])
        >>> euclideanDistance(x, y)
        '''
        x: np.array = np.array(x)
        y: np.array = np.array(y)
        x_expanded: np.array = x[:, np.newaxis, :]
        y_expanded: np.array = y[np.newaxis, :, :]
        return np.sqrt(np.sum((x_expanded - y_expanded) ** 2, axis=2))
    
    @staticmethod
    def cosineSimilarity(x: np.array,
                         y: np.array) -> np.array:
        '''
        method cosineSimilarity: this method is used to compute the cosine similarity between two matrices

        Remark: this method is an obtimization of the cosine similarity computation between two matrices

        : param x: np.array the first matrix
        : param y: np.array the second matrix
        : return: np.array the cosine similarity matrix between the two matrices

        Usage example:
        >>> x: np.array = np.array([1, 2, 3])
        >>> y: np.array = np.array([4, 5, 6])
        >>> cosineSimilarity(x, y)
        '''
        x: np.array = np.array(x)
        y: np.array = np.array(y)
        x_expanded: np.array = x[:, np.newaxis, :]
        y_expanded: np.array = y[np.newaxis, :, :]
        return np.sum(x_expanded * y_expanded, axis=2) / (np.linalg.norm(x_expanded, axis=2) * np.linalg.norm(y_expanded, axis=2))


import numpy as np
import numpy.random as npr
from .AbstractEmbeddingPerturbationMechanism import AbstractEmbeddingPerturbationMechanism
from typing import List

'''
# CMP Mechanism
'''

class CMP(AbstractEmbeddingPerturbationMechanism):
    '''
    # CMP
    
    BibTeX of CMP Mechanism, extends Mechanism mechanism class of the pypantera package:

    @inproceedings{FeyisetanEtAl2020CMP,
    author = {Feyisetan, Oluwaseyi and Balle, Borja and Drake, Thomas and Diethe, Tom},
    title = {Privacy- and Utility-Preserving Textual Analysis via Calibrated Multivariate Perturbations},
    year = {2020},
    isbn = {9781450368223},
    publisher = {Association for Computing Machinery},
    address = {New York, NY, USA},
    url = {https://doi.org/10.1145/3336191.3371856},
    doi = {10.1145/3336191.3371856},
    abstract = {Accurately learning from user data while providing quantifiable privacy guarantees provides an opportunity to build better ML models while maintaining user trust. This paper presents a formal approach to carrying out privacy preserving text perturbation using the notion of d_χ-privacy designed to achieve geo-indistinguishability in location data. Our approach applies carefully calibrated noise to vector representation of words in a high dimension space as defined by word embedding models. We present a privacy proof that satisfies d_χ-privacy where the privacy parameter $varepsilon$ provides guarantees with respect to a distance metric defined by the word embedding space. We demonstrate how $varepsilon$ can be selected by analyzing plausible deniability statistics backed up by large scale analysis on GloVe and fastText embeddings. We conduct privacy audit experiments against $2$ baseline models and utility experiments on 3 datasets to demonstrate the tradeoff between privacy and utility for varying values of varepsilon on different task types. Our results demonstrate practical utility (< 2\% utility loss for training binary classifiers) while providing better privacy guarantees than baseline models.},
    booktitle = {Proceedings of the 13th International Conference on Web Search and Data Mining},
    pages = {178-186},
    numpages = {9},
    keywords = {privacy, plausible deniability, differential privacy},
    location = {Houston, TX, USA},
    series = {WSDM '20}
    }
    '''
    
    def __init__(self, kwargs: dict[str:object]) -> None:

        '''
        Initialization of the Mechanism Object

        : param kwargs: dict the dictionary containing the parameters of the Mechanism Object

        Once the Mechanism Object is created, the user can use the obfuscateText method to obfuscate 
        the text of the provided text.

        The attributes of the Mechanism Object are:
        - vocab: the vocabulary object containing the embeddings
        - embMatrix: the matrix containing the embeddings
        - index2word: the dictionary containing the index to word mapping
        - word2index: the dictionary containing the word to index mapping
        - epsilon: the epsilon parameter of the mechanism

        Usage example:
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1 #anyvalue of epsilon must be greater than 0
        >>> mech1 = Mechanism({'embPath': embPath, 'epsilon': eps})
        '''
        super().__init__(kwargs)
        self.name: str = 'CMP'
        

    def pullNoise(self) -> np.array:
        '''
        method pullNoise: this method is used to pull noise accordingly 
        to the definition of the CMP mechanism, see BibTeX ref

        : return: np.array the noise pulled

        Usage example:
        (Considering that the Mechanism Object mech1 has been created 
        as in the example of the __init__ method)
        >>> mech1.pullNoise()
        '''

        N: np.array = self.epsilon * npr.multivariate_normal(
            np.zeros(self.embMatrix.shape[1]),
            np.eye(self.embMatrix.shape[1])) #pull noise from a multivariate normal distribution
        X: np.array = N / np.sqrt(np.sum(N ** 2)) #normalize the noise
        Y: np.array = npr.gamma(
            self.embMatrix.shape[1],
            1 / self.epsilon) #pull gamma noise
        Z: np.array = Y * X #compute the final noise
        return Z

    def processText(self, 
                     embs: np.array) -> str:
        
        '''
        method processText: this method is used to process the Text and return the obfuscated Text

        : param embs: np.array the embeddings of the words
        : return: str the obfuscated Text

        Usage example:
        (Considering that the Mechanism Object mech1 has been created
        as in the example of the __init__ method)

        # Assuming that the embeddings of the words are known, e.g.: [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        >>> embs: np.array = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        >>> mech1.processText(embs)
        '''
        length: int = len(embs)
        distance: np.array = self.euclideanDistance(embs, self.embMatrix)
        closest: np.array = np.argpartition(distance, 1, axis=1)[:, :1]
        finalText: List[str] = []
        for i in range(length):
            finalText.append(list(self.vocab.embeddings.keys())[closest[i][0]])
        return ' '.join(finalText)


from ..AbstractTextObfuscationDPMechanism import AbstractTextObfuscationDPMechanism
import numpy as np
import pandas as pd
from typing import List
import numpy.random as npr

class AbstractEmbeddingPerturbationMechanism(AbstractTextObfuscationDPMechanism):
    '''
    Abstract class scramblingEmbeddingsMechanism: this class is used to define the abstract class of the scrambling Embeddings mechanisms
    '''
    def __init__(self, kwargs: dict[str:object]) -> None:
        '''
        Constructur of the scramblingEmbeddingsMechanism class
        '''
        super().__init__(kwargs)
    
    def noisyEmb(self, words: List[str]) -> np.array:
        '''
        method noisyEmb: this method is used to add noise to the embeddings of the words

        : param words: List[str] the list of words to add noise to
        : return: np.array the noisy embeddings

        Usage example:
        (Considering that the Mechanism Object mech1 has been created
        as in the example of the __init__ method)
        >>> words: List[str] = ['what', 'is', 'the', 'capitol', 'of', 'france']
        >>> mech1.noisyEmb(words)
        '''
        embs: List = []
        for word in words:
            if word not in self.vocab.embeddings:
                embs.append(
                    np.zeros(self.embMatrix.shape[1]) + npr.normal(0, 1, self.embMatrix.shape[1]) #handle OoV words
                    + self.pullNoise()
                    )
            else:
                embs.append(self.vocab.embeddings[word] + self.pullNoise())
        return np.array(embs)
    
    def obfuscateText(self, data:pd.DataFrame, numberOfTimes: int) -> List[str]:
        '''
        method obfuscateText: this method is used to obfuscate the text of the provided text 
        using the Mahalanobis mechanism

        : param data: str the text to obfuscate
        : param numberOfCores: int the number of cores to use for the obfuscation
        : return: str the obfuscated text

        Usage example:
        (Considering that the Mechanism Object mech1 has been created
        as in the example of the __init__ method)

        >>> text: str = 'what is the capitol of france'
        >>> mech1.obfuscateText(text, 1)
        '''
        #rename the columns
        data.columns = ['id', 'text']
        #tokenize the text column
        data['text'] = data['text'].apply(lambda x: x.lower().split())
        df = pd.DataFrame(columns=['id', 'text', 'obfuscatedText'])
        #obfuscate the text column
        obfuscatedText = pd.concat([data['text'].apply(lambda x: self.processText(self.noisyEmb(x))) for _ in range(numberOfTimes)], axis=1)#note that here we're passing the embeddings noisy
        df = pd.concat([data, pd.DataFrame({'obfuscatedText': obfuscatedText.values.tolist()})], axis=1)
        df['text'] = df['text'].apply(lambda x: ' '.join(x))
        return df


from .AbstractEmbeddingPerturbationMechanism import AbstractEmbeddingPerturbationMechanism
from .cmp import CMP
from .mahalanobis import Mahalanobis
import numpy as np
import numpy.random as npr
from typing import List

class AbstractVickrey(AbstractEmbeddingPerturbationMechanism):
    '''
    Vickrey Mechanism
    '''
    def __init__(self, kwargs: dict[str:object]) -> None:
        '''
        # Vickrey
        BibTeX of Vickrey Mechanism, extends Mechanism class of the pypanter package:

        @inproceedings{xu-etal-2021-utilitarian,
            title = "On a Utilitarian Approach to Privacy Preserving Text Generation",
            author = "Xu, Zekun  and
              Aggarwal, Abhinav  and
              Feyisetan, Oluwaseyi  and
              Teissier, Nathanael",
            editor = "Feyisetan, Oluwaseyi  and
              Ghanavati, Sepideh  and
              Malmasi, Shervin  and
              Thaine, Patricia",
            booktitle = "Proceedings of the Third Workshop on Privacy in Natural Language Processing",
            month = jun,
            year = "2021",
            address = "Online",
            publisher = "Association for Computational Linguistics",
            url = "https://aclanthology.org/2021.privatenlp-1.2",
            doi = "10.18653/v1/2021.privatenlp-1.2",
            pages = "11--20",
            abstract = "Differentially-private mechanisms for text generation typically add carefully calibrated noise to input words and use the nearest neighbor to the noised input as the output word. When the noise is small in magnitude, these mechanisms are susceptible to reconstruction of the original sensitive text. This is because the nearest neighbor to the noised input is likely to be the original input. To mitigate this empirical privacy risk, we propose a novel class of differentially private mechanisms that parameterizes the nearest neighbor selection criterion in traditional mechanisms. Motivated by Vickrey auction, where only the second highest price is revealed and the highest price is kept private, we balance the choice between the first and the second nearest neighbors in the proposed class of mechanisms using a tuning parameter. This parameter is selected by empirically solving a constrained optimization problem for maximizing utility, while maintaining the desired privacy guarantees. We argue that this empirical measurement framework can be used to align different mechanisms along a common benchmark for their privacy-utility tradeoff, particularly when different distance metrics are used to calibrate the amount of noise added. Our experiments on real text classification datasets show up to 50{\%} improvement in utility compared to the existing state-of-the-art with the same empirical privacy guarantee.",
        }
        '''
        super().__init__(kwargs)
        assert 't' in kwargs, 'The t parameter must be provided and between 0 and 1'
        assert kwargs['t'] >= 0 and kwargs['t'] <= 1, 'The t parameter must be between 0 and 1'
        self.t: float = kwargs['t']
        self.name: str = 'Vickrey'

class VickreyCMP(CMP, AbstractVickrey):
    def __init__(self, kwargs) -> None:
           
            '''
            Initialization of the VickreyCMP Object
            : param kwargs: dict the dictionary containing the parameters of the Mechanism Object 
                            + the specific parameters of the VickreyCMP Mechanism (lambda)
            Once the VickreyCMP Object is created, the user can use the obfuscateText method to obfuscate 
            the text of the provided text.
            The attributes of the Mechanism Object are:
            - vocab: the vocabulary object containing the embeddings
            - embMatrix: the matrix containing the embeddings
            - index2word: the dictionary containing the index to word mapping
            - word2index: the dictionary containing the word to index mapping
            - epsilon: the epsilon parameter of the mechanism
            - t: the t parameter of the Vickrey mechanism
            Usage example:
            >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
            >>> eps: float = 0.1 #anyvalue of epsilon must be greater than 0
            >>> t: float = 0.5 #anyvalue of t must be between 0 and 1
            >>> mech1 = Vickrey.Mhl({'embPath': embPath, 'epsilon': eps, 't': t})
            '''
            super().__init__(kwargs)
            self.name:str = 'VickreyCMP'

    def processText(self, 
                 embs: np.array) -> str:
        '''
        method processText: this method is used to process the Text accordingly
        to the definition of the Vickrey mechanism, see BibTeX ref

        : param embs: np.array the embeddings of the Text
        : return: str the obfuscated Text

        Usage example:
        (Considering that the Mechanism Object mech1 has been created
        as in the example of the __init__ method)
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1
        >>> lam: float = 0.1
        >>> t: float = 0.5

        >>> mech1 = Vickrey.Mhl({'embPath': embPath, 'epsilon': eps, 'lambda': lam, 't': t})
        >>> Text: str = 'what is the capitol of france'
        >>> embs: np.array = mech1.getEmbeddings(Text)
        >>> obfuscatedText: str = mech1.processText(embs)
        '''

        length: int = len(embs) #compute number of words

        distance: np.array = self.euclideanDistance(embs, self.embMatrix) #compute distances between words and embeddings in the vocabulary
        closest: np.array = np.argpartition(distance, 2, axis=1)[:, :2] #find the two closest embeddings for each word
        
        distToClosest: np.array = distance[np.tile(np.arange(length).reshape(-1,1),2), closest] #compute the distances to the two closest embeddings
        p = ((1- self.t) * distToClosest[:,1]) / (self.t * distToClosest[:,0] + (1 - self.t) * distToClosest[:, 1]) #compute the probabilities of choosing the second closest embedding
        vickreyChoice: np.array = np.array([npr.choice(2, p=[p[w], 1-p[w]]) for w in range(length)]) #choose the closest embedding according to the probabilities
        noisyEmbeddings: np.array = self.embMatrix[closest[np.arange(length), vickreyChoice]] #get the noisy embeddings

        distance = self.euclideanDistance(noisyEmbeddings, self.embMatrix)
        found = np.argpartition(distance, 1, axis=1)[:, :1]
        
        finalText = self.indexes2words(found)       
        return ' '.join(finalText)
        
    
class VickreyMhl(Mahalanobis, AbstractVickrey):
    def __init__(self, kwargs) -> None:
        '''
        Initialization of the VickreyMahalanobis Object 
        : param kwargs: dict the dictionary containing the parameters of the Mechanism Object 
                        + the specific parameters of the VickreyMahalanobis Mechanism (lambda)  

        Once the VickreyMahalanobis Object is created, the user can use the obfuscateText method to obfuscate 
        the text of the provided text.  

        The attributes of the Mechanism Object are:
        - vocab: the vocabulary object containing the embeddings
        - embMatrix: the matrix containing the embeddings
        - index2word: the dictionary containing the index to word mapping
        - word2index: the dictionary containing the word to index mapping
        - epsilon: the epsilon parameter of the mechanism
        - lam: the lambda parameter of the Mahalanobis mechanism
        - sigma_loc: parameter used for pulling noise in obfuscation
        - t: the t parameter of the Vickrey mechanism   

        Usage example:
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1 #anyvalue of epsilon must be greater than 0
        >>> lam: float = 0.1 #anyvalue of lambda must be between 0 and 1
        >>> t: float = 0.5 #anyvalue of t must be between 0 and 1
        >>> mech1 = Vickrey.Mhl({'embPath': embPath, 'epsilon': eps, 'lambda': lam, 't': t})
        '''
        super().__init__(kwargs)
        self.name:str = 'VickreyMhl'

    def processText(self, 
                 embs: np.array) -> str:
        '''
        method processText: this method is used to process the Text accordingly
        to the definition of the Vickrey mechanism, see BibTeX ref

        : param embs: np.array the embeddings of the Text
        : return: str the obfuscated Text

        Usage example:
        (Considering that the Mechanism Object mech1 has been created
        as in the example of the __init__ method)
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1
        >>> lam: float = 0.1
        >>> t: float = 0.5

        >>> mech1 = Vickrey.Mhl({'embPath': embPath, 'epsilon': eps, 'lambda': lam, 't': t})
        >>> text: str = 'what is the capitol of france'
        >>> embs: np.array = mech1.getEmbeddings(text)
        >>> obfuscatedText: str = mech1.processText(embs)
        '''

        length: int = len(embs) #compute number of words

        distance: np.array = self.euclideanDistance(embs, self.embMatrix) #compute distances between words and embeddings in the vocabulary
        closest: np.array = np.argpartition(distance, 2, axis=1)[:, :2] #find the two closest embeddings for each word
        
        distToClosest: np.array = distance[np.tile(np.arange(length).reshape(-1,1),2), closest] #compute the distances to the two closest embeddings
        p = ((1- self.t) * distToClosest[:,1]) / (self.t * distToClosest[:,0] + (1 - self.t) * distToClosest[:, 1]) #compute the probabilities of choosing the second closest embedding
        vickreyChoice: np.array = np.array([npr.choice(2, p=[p[w], 1-p[w]]) for w in range(length)]) #choose the closest embedding according to the probabilities
        noisyEmbeddings: np.array = self.embMatrix[closest[np.arange(length), vickreyChoice]] #get the noisy embeddings
        
        distance = self.euclideanDistance(noisyEmbeddings, self.embMatrix)
        found = np.argpartition(distance, 1, axis=1)[:, :1]
        
        finalText = self.indexes2words(found)
       
        return ' '.join(finalText)
        


import os
import logging
import pandas as pd
import argparse
from typing import List

from ..AbstractTextObfuscationDPMechanism import AbstractTextObfuscationDPMechanism
#from ..EmbeddingPerturbationMechanism import AbstractEmbeddingPerturbationMechanism
#from ..EmbeddingPerturbationMechanism import AbstractSamplingPerturbationMechanism

from ..EmbeddingPerturbationMechanism.cmp import CMP
from ..EmbeddingPerturbationMechanism.mahalanobis import Mahalanobis
from ..EmbeddingPerturbationMechanism.vickrey import VickreyCMP, VickreyMhl

from ..SamplingPerturbationMechanism.santext import SanText
from ..SamplingPerturbationMechanism.custext import CusText
from ..SamplingPerturbationMechanism.tem import TEM

'''
helper module: Provides helper functions to create the parser, the logger and to select the mechanism to use


'''

def createParser() -> object:
    '''
    # createParser()
    creates the parser object and returns it
    The arguments should be passed by command line otherwise the program will crush if no embPath and inputPath are provided

    # Required Params:
    - embPath: the path to the embeddings file
    - inputPath: the path to the input file (.csv file)

    # Default Params :
    - mechanism: the mechanism to use (default: CMP)
    - task: the task to compleate
    - t: the treshold value to use Vickrey mechanisms (default: 0.75)
    - beta: the beta value to use for the TEM mechanism (default: 0.001)
    - lam: the lambda value to use for the Mahalanobis and Vickrey mechanisms (default: 1)
    - k: the number of words to sample for the CusText mechanism (default: 5)
    - epsilons: the list of epsilon values to use (default: [1, 5, 10, 12.5, 15, 17.5, 20, 50])
    - numberOfObfuscations: the number of obfuscations to perform (default: 1)
    '''
    parser:object = argparse.ArgumentParser(description='Obfuscate texts using different mechanisms')

    #mechanism params
    parser.add_argument('--task', '-tk', type=str, help='The task to perform', default='retrieval')
    parser.add_argument('--embPath', '-eP', type=str, help='The path to the embeddings file')
    parser.add_argument('--inputPath', '-i', type=str, help='The path to the input file')
    parser.add_argument('--outputPath', '-o', type=str, help='The path to the output file')
    parser.add_argument('--mechanism', '-m', type=str, help='The mechanism to use', default='CMP', choices=['CMP', 'Mahalanobis', 'VickreyCMP', 'VickreyMhl', 'CusText', 'SanText', 'TEM'])
    parser.add_argument('--t', '-t', type=float, help='The treshold value to use Vickrey mechanisms', default=0.75)
    parser.add_argument('--beta', '-beta', type=float, help='The beta value to use for the TEM mechanism', default=0.001)
    parser.add_argument('--lam', '-lam', type=float, help='The lambda value to use for the Mahalanobis and Vickrey mechanisms', default=1)
    parser.add_argument('--k', '-k', type=int, help='The number of words to sample for the CusText mechanism', default=10)
    parser.add_argument('--distance', '-d', type=str, help='The distance metric to use for the CusText mechanism', default='euclidean')
    parser.add_argument('--epsilons','-e', type=float, help='The list of epsilon values to use', nargs='+', default=[1, 5, 10, 12.5, 15, 17.5, 20, 50])
    parser.add_argument('--numberOfObfuscations', '-n', type=int, help='The number of obfuscations to perform', default=1)
    return parser

def createLogger() -> object:
    '''
    # createLogger()
    creates the logger object and returns it
    Runtime logs can be found in the pypantera/logs/logger.log file
    '''
    if os.path.exists('./pypantera/logs/logger.log'):
        os.remove('./pypantera/logs/logger.log')
    else:
        os.mkdir('./pypantera/logs')
    
    FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    logging.basicConfig(filename='./pypantera/logs/logger.log', 
                        level=logging.INFO,
                        format=FORMAT)
    logger = logging.getLogger(__name__)
    return logger

def selectMechanism(args:object, logger:object) -> list:
    '''
    select the mechanism to use based on the argument.mechanism provided, initialize a list of mechanisms for different parameters
    '''
    if args.mechanism in ['CMP', 'Mahalanobis', 'VickreyCMP', 'VickreyMhl']:
        if args.mechanism == 'CMP':
            mechanisms = [CMP({'embPath': args.embPath, 'epsilon': e}) for e in args.epsilons]
        elif args.mechanism == 'Mahalanobis':
            mechanisms = [Mahalanobis({'embPath': args.embPath, 'epsilon': e, 'lam': args.lam}) for e in args.epsilons]
            logger.info(f"Lambda: {mechanisms[0].lam}")
        elif args.mechanism == 'VickreyCMP':
            mechanisms = [VickreyCMP({'embPath': args.embPath, 'epsilon': e, 't':args.t}) for e in args.epsilons]
            logger.info(f"Treshold: {mechanisms[0].t}")
        elif args.mechanism == 'VickreyMhl':
            mechanisms = [VickreyMhl({'embPath': args.embPath, 'epsilon': e, 'lam': args.lam, 't': args.t}) for e in args.epsilons]
            logger.info(f"Lambda: {mechanisms[0].lam}")
            logger.info(f"Treshold: {mechanisms[0].t}")
    elif args.mechanism in ['SanText', 'CusText', 'TEM']:
        if args.mechanism == 'SanText':
            mechanisms = [SanText({'embPath': args.embPath, 'epsilon': e}) for e in args.epsilons]
        elif args.mechanism == 'CusText':
            mechanisms = [CusText({'embPath': args.embPath, 'epsilon': e, 'k':args.k, 'distance':args.distance}) for e in args.epsilons]
            logger.info(f"K: {mechanisms[0].k}")
            logger.info(f"Distance: {mechanisms[0].distance}")
        elif args.mechanism == 'TEM':
            mechanisms = [TEM({'embPath': args.embPath, 'epsilon': e, 'beta': args.beta}) for e in args.epsilons]
            logger.info(f"Beta: {mechanisms[0].beta}")
            for i in range(len(mechanisms)):
                logger.info(f"For mechansim with epsilon = {mechanisms[i].epsilon}, Gamma computed: {mechanisms[i].gamma}")
    else:
        logger.error('The mechanism provided is not valid. Please choose one of the following mechanisms: CMP, Mahalanobis, VickreyCMP, VickreyMhl, CusText')
        exit(1)
    logger.info(f"Initialized one mechanism of type {mechanisms[0].__class__.__name__} for each epsilon value")
    return mechanisms

def saveResults(results:List[pd.DataFrame], mechanisms:List[AbstractTextObfuscationDPMechanism], args:object, logger:object, path:str = None) -> None:
    '''
    Save results to a csv file in the results/args.task folder
    '''
    for i, df in enumerate(results):
        df:pd.DataFrame = df.explode('obfuscatedText')
        df['mechanism'] = mechanisms[i].__class__.__name__
        df['epsilon'] = mechanisms[i].epsilon
        #save the obfuscated text to a csv file
        output_path = args.outputPath
        if path == None:
            if not os.path.exists('./results'):
                os.makedirs('./results')
            if not os.path.exists(f'./results/{args.task}'):
                os.makedirs(f'./results/{args.task}')
            if not os.path.exists(f'./results/{args.task}/{mechanisms[i].__class__.__name__}'):
                os.makedirs(f'./results/{args.task}/{mechanisms[i].__class__.__name__}')
            df.to_csv(f'./results/{args.task}/{mechanisms[i].__class__.__name__}/obfuscatedText_{mechanisms[i].__class__.__name__}_{mechanisms[i].epsilon}.csv', index=False)
            logger.info(f"Saved the obfuscated queries to a csv file for mechanism {mechanisms[i].__class__.__name__} with epsilon {mechanisms[i].epsilon}")
        else:
            df.to_csv(path, index=False)
            logger.info(f"Saved the obfuscated queries to a csv file for mechanism {mechanisms[i].__class__.__name__} with epsilon {mechanisms[i].epsilon}")
    logger.info('Obfuscation process completed successfully!')
    os.system(f'cp ./pypantera/logs/logger.log ./results/{args.task}/{mechanisms[0].__class__.__name__}/logger_{mechanisms[0].__class__.__name__}.log')
        


import os
import numpy as np
import multiprocessing as mp

class Vocab():
    def __init__(self, embPath: str) -> None:
        self.loadParallel(embPath)

    def loadParallel(self, embPath: str) -> None:
        if not os.path.exists(embPath):
            raise FileNotFoundError(f"File {embPath} not found")

        # Determine the number of processes based on the available CPUs
        num_processes = mp.cpu_count()

        # Determine the size of each chunk to be processed by each process
        fileSize = os.path.getsize(embPath)
        chunkSize = fileSize // num_processes

        # Create a pool of processes
        with mp.Pool(processes=num_processes) as pool:
            # Map each chunk to a separate process
            results = pool.map(func=self.loadChunk, iterable=[(embPath, i * chunkSize, (i+1) * chunkSize) for i in range(num_processes)])

        # Combine the results from all processes
        self.embeddings = {}
        for result in results:
            self.embeddings.update(result)

    def loadChunk(self, args) -> dict:
        embPath, start, end = args
        embeddings = {}

        with open(embPath, 'r') as file:
            if start != 0:
                file.seek(start)
                file.readline()  # Skip the first line as it may be incomplete
            while file.tell() < end:
                line = file.readline()
                if not line:
                    break  # End of file
                values = line.split()
                word = values[0]
                vector = np.asarray(values[1:], "float32")
                embeddings[word] = vector

        return embeddings



from .AbstractSamplingPerturbationMechanism import AbstractSamplingPerturbationMechanism
import numpy as np
import numpy.random as npr
from scipy.linalg import sqrtm
import multiprocessing as mp
from typing import List
import time
from collections import defaultdict
from sklearn.metrics.pairwise import euclidean_distances

'''
# CusText Mechanism
'''

class CusText(AbstractSamplingPerturbationMechanism):
    '''
    # CusText

    BibTeX of CusText Mechanism, extends CMP mechanism class of the pypanter package:

    @inproceedings{ChenEtAl2023Customized,
    title = {A Customized Text Sanitization Mechanism with Differential Privacy},
    author = {Chen, Sai and Mo, Fengran and Wang, Yanhao and Chen, Cen and Nie, Jian-Yun and Wang, Chengyu and Cui, Jamie},
    editor = {Rogers, Anna and Boyd-Graber, Jordan and Okazaki, Naoaki},
    booktitle = {Findings of the Association for Computational Linguistics: ACL 2023},
    month = jul,
    year = {2023},
    address = {Toronto, Canada},
    publisher = {Association for Computational Linguistics},
    url = {https://aclanthology.org/2023.findings-acl.355},
    doi = {10.18653/v1/2023.findings-acl.355},
    pages = {5747--5758},
    }    
    '''
    def __init__(self, kwargs: dict[str:object]) -> None:
        '''
        Initialization of the CusText Object

        : param kwargs: dict the dictionary containing the parameters of the Mechanism Object 
                        + the specific parameters of the CusText Mechanism ()

        Once the CusText Object is created, the user can use the obfuscateText method to obfuscate 
        the text of the provided text.

        The attributes of the Mechanism Object are:
        - vocab: the vocabulary object containing the embeddings
        - embMatrix: the matrix containing the embeddings
        - index2word: the dictionary containing the index to word mapping
        - word2index: the dictionary containing the word to index mapping
        - epsilon: the epsilon parameter of the mechanism
        - 

        Usage example:
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1 #anyvalue of epsilon must be greater than 0
        >>> k: int = 5 #any integer greater than 0 
        >>> mech1 = CusText({'embPath': embPath, 'epsilon': eps, })
        '''
        super().__init__(kwargs)
        assert 'k' in kwargs, 'The k parameter must be provided'
        assert kwargs['k'] > 0, 'The k parameter must be greater than 0'
        self.k: int = kwargs['k']
        assert 'distance' in kwargs, 'The distance parameter must be provided'
        assert kwargs['distance'] in ['euclidean', 'cosine'], 'The distance parameter must be either euclidean or cosine'
        self.distance:str = kwargs['distance']
        self.name:str = 'CusText'
        self._simWordDict = {}
        self._pDict = {}

    def _getCustomizedMapping(self, word) -> tuple[defaultdict, defaultdict]:
        '''
        precompute the mapping of the words and their respective probabilities
        '''
        simWordDict:defaultdict = defaultdict(list)
        pDict:defaultdict = defaultdict(list)
        wordFreq = list(self._simWordDict.keys())
        if self.distance == 'euclidean':
            if word not in wordFreq:
                try:
                    wordEmb = self.vocab.embeddings[word].reshape(1,-1)
                except:
                    wordEmb = np.zeros(self.embMatrix.shape[1]) + npr.normal(0, 1, self.embMatrix.shape[1]).reshape(1,-1)
                    
                indexList = self.euclideanDistance(wordEmb, self.embMatrix).argsort()[:self.k][0]
                wordList = [self._index2word[index] for index in indexList[:self.k]]
                embeddingList = np.array([self.vocab.embeddings[w] for w in wordList])

                simDistList = self.euclideanDistance(wordEmb, embeddingList)[0]
                min_max_dist = max(simDistList) - min(simDistList)
                min_dist = min(simDistList)
                newSimDistList = [-(x-min_dist)/min_max_dist for x in simDistList]
                tmp = [np.exp(self.epsilon*x/2) for x in newSimDistList]
                norm = sum(tmp)
                p = [x/norm for x in tmp]
                pDict[word] = p
                simWordDict[word] = wordList
                self._pDict[word] = p
                self._simWordDict[word] = wordList 
            else:
                simWordDict[word] = self._simWordDict[word]
                pDict[word] = self._pDict[word]

        elif self.distance == 'cosine':
            if word not in wordFreq:
                indexList = self.cosineSimilarity(self.vocab.embeddings[word].reshape(1,-1), self.embMatrix).argsort()[:self.k][0]
                wordList = [self._index2word[index] for index in indexList[:self.k]]
                embeddingList = np.array([self.vocab.embeddings[w] for w in wordList])

                simDistList = self.cosineSimilarity(self.vocab.embeddings[word].reshape(1,-1), embeddingList)[0]
                min_max_dist = max(simDistList) - min(simDistList)
                min_dist = min(simDistList)
                newSimDistList = [(x-min_dist)/min_max_dist for x in simDistList]
                tmp = [np.exp(self.epsilon*x/2) for x in newSimDistList]
                norm = sum(tmp)
                p = [x/norm for x in tmp]
                pDict[word] = p
                simWordDict[word] = wordList
                self._pDict[word] = p
                self._simWordDict[word] = wordList 
            else:
                simWordDict[word] = self._simWordDict[word]
                pDict[word] = self._pDict[word]
        return simWordDict, pDict
        ...    

    #def _generateText(self, text:str) -> str:
    #    ...

    def processText(self, text:List[str])->str:
        ''' 
        method selfProcessQueryText: this method is used to process the Text and return the obfuscated Text

        : param embs: np.array the embeddings of the words
        : return: str the obfuscated Text
        '''
        finalText:list = []
        for word in text:
            simWordDict, pDict = self._getCustomizedMapping(word)
            newWord:str = np.random.choice(simWordDict[word], 1, p=pDict[word])[0]
            finalText.append(newWord)
        return ' '.join(finalText)
            
        ...
        



            




from .AbstractSamplingPerturbationMechanism import AbstractSamplingPerturbationMechanism
import numpy as np
import numpy.random as npr
from scipy.linalg import sqrtm
import multiprocessing as mp
from typing import List

class TEM(AbstractSamplingPerturbationMechanism):
    '''
    BibTeX of TEM Mechanism, extends CMP mechanism class of the pypanter package:

    @inbook{doi:10.1137/1.9781611977653.ch99,
        author = {Ricardo Silva Carvalho and Theodore Vasiloudis and Oluwaseyi Feyisetan and Ke Wang},
        title = {TEM: High Utility Metric Differential Privacy on Text},
        booktitle = {Proceedings of the 2023 SIAM International Conference on Data Mining (SDM)},
        chapter = {},
        pages = {883-890},
        doi = {10.1137/1.9781611977653.ch99},
        URL = {https://epubs.siam.org/doi/abs/10.1137/1.9781611977653.ch99},
        eprint = {https://epubs.siam.org/doi/pdf/10.1137/1.9781611977653.ch99},
            abstract = { Abstract Ensuring the privacy of users whose data are used to train Natural Language Processing (NLP) models is necessary to build and maintain customer trust. Differential Privacy (DP) has emerged as the most successful method to protect the privacy of individuals. However, applying DP to the NLP domain comes with unique challenges. The most successful previous methods use a generalization of DP for metric spaces, and apply the privatization by adding noise to inputs in the metric space of word embeddings. However, these methods assume that one specific distance measure is being used, ignore the density of the space around the input, and assume the embeddings used have been trained on public data. In this work we propose Truncated Exponential Mechanism (TEM), a general method that allows the privatization of words using any distance metric, on embeddings that can be trained on sensitive data. Our method makes use of the exponential mechanism to turn the privatization step into a selection problem. This allows the noise applied to be calibrated to the density of the embedding space around the input, and makes domain adaptation possible for the embeddings. In our experiments, we demonstrate that our method outperforms the state-of-the-art in terms of utility for the same level of privacy, while providing more flexibility in the metric selection. }
        }
    '''
    def __init__(self, kwargs: dict[str:object]) -> None:
        '''
        Initialization of the TEM Object

        : param kwargs: dict the dictionary containing the parameters of the Mechanism Object 
                        + the specific parameters of the TEM Mechanism ()

        Once the TEM Object is created, the user can use the obfuscateText method to obfuscate 
        the text of the provided text.

        The attributes of the Mechanism Object are:
        - vocab: the vocabulary object containing the embeddings
        - embMatrix: the matrix containing the embeddings
        - index2word: the dictionary containing the index to word mapping
        - word2index: the dictionary containing the word to index mapping
        - epsilon: the epsilon parameter of the mechanism
        - 

        Usage example:
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1 #anyvalue of epsilon must be greater than 0
        >>> beta: float = 0.001 #anyvalue of beta must be greater than 0 and less than 1
        >>> mech1 = TEM({'embPath': embPath, 'epsilon': eps, })
        '''
        super().__init__(kwargs)
        assert 'beta' in kwargs, 'The beta parameter must be provided'
        assert kwargs['beta'] > 0 and kwargs['beta'] < 1, 'The beta parameter must be between 0 and 1'
        self.beta: float = kwargs['beta'] 
        self.gamma: float = (2/self.epsilon)*np.log(((1-self.beta)*(len(self.embMatrix)-1))/self.beta)
        self.name:str = 'TEM'

        self.candidates: dict = {}
        #self._internalPreprocessing()

    def pullNoise(self) -> np.array:
        '''
        pullNoise method: this method is used to pull noise from the Laplace distribution
        : param n: int the number of noise to pull (size of the Lw precomputed list)

        : return: np.array the noise pulled from the Laplace distribution

        Usage example:
        (Considering that the Mechanism Object mech1 has been created
        as in the example of the __init__ method)
        >>> mech1.pullNoise()
        '''
        gumbel_mean: float = 0
        gumbel_scale: float = 2/self.epsilon
        return npr.gumbel(gumbel_mean, gumbel_scale, 1)

    def _getLw(self, word:str) -> tuple[List[tuple], List[str]]:
        distance:np.array = self.euclideanDistance(self.vocab.embeddings[word].reshape(1,-1), self.embMatrix)
        #select only distances that are below self.gamma
        indices_Lw:list = np.where(distance[0] <= self.gamma)[0]
        indices_L_hat_w:list = np.where(distance[0] > self.gamma)[0]
        Lw:List[tuple] = [(self._index2word[i], -distance[0][i]) for i in indices_Lw]
        L_hat_w:list = [self.embMatrix[i] for i in indices_L_hat_w]
        if len(L_hat_w) == 0:
            #set value to -inf
            temp_score:float = -np.inf
        else:
            temp_score:float = -self.gamma + 2 *np.log(len(L_hat_w))/self.epsilon
        tempWord:str = 'PLACEHOLDERWORD'
        Lw.append((tempWord, temp_score))
        return Lw, L_hat_w

    def processText(self, 
                 text:List[str]) -> str:
        '''
        processText method: this method is used to process the Text accordingly
        to the definition of the TEM mechanism, see BibTeX ref

        : param embs: np.array the embeddings of the Text
        : return: str the obfuscated Text

        Usage example:
        >>> embPath: str = 'pathToMyEmbeddingsFile.txt'
        >>> eps: float = 0.1
        >>> beta: float = 0.001
        >>> mech1 = TEM({'embPath': embPath, 'epsilon': eps, 'beta': beta})
        >>> words: List[str] = ['what is the capitol of france']
        >>> embs: np.array = mech1.getEmbeddings(words)
        >>> obfuscatedText: str = mech1.processText(embs)
        '''
        #print(f'text: {text}')
        def _processWord(word:str) -> str:
            '''
            _processWord method: this method is used to process the word accordingly to the definition of the TEM mechanism

            : param word: str the word to process
            : return: str the processed obfuscated word
            '''
            if word not in self.vocab.embeddings.keys():
                self.vocab.embeddings[word] = np.zeros(self.embMatrix.shape[1]) + npr.normal(0, 1, self.embMatrix.shape[1]) #handle OoV words
            Lw, L_hat_w = self._getLw(word)
            #add to the scores in Lw the noise
            Lw:List[tuple]= [(w, s + self.pullNoise()) for w, s in Lw]
            #get the word with the highest score
            selectedWord:str = max(Lw, key=lambda x: x[1])[0]
            if selectedWord == 'PLACEHOLDERWORD':
                try:
                    selectedWord:str = np.random.choice(L_hat_w, 1)[0]
                except:
                    selectedWord:str = word
            return selectedWord
        
        finalText:List[str] = list(map(_processWord, text))
        return ' '.join(finalText)
            

        




        




        ...